from pathlib import Path
from typing import Optional

import faiss
import numpy as np
import pytest

from semsis.retriever.faiss_cpu import RetrieverFaissCPU, faiss_index_builder

N = 3
D = 8
M = 4
nbits = 4
nlists = 4


@pytest.fixture(scope="module")
def v() -> np.ndarray:
    np.random.seed(0)
    return np.random.rand(N, D).astype(np.float32)


@pytest.mark.parametrize("hnsw_nlinks", [0, 4])
@pytest.mark.parametrize("pq_nblocks", [0, M])
@pytest.mark.parametrize("ivf_nlists", [0, nlists])
def test_faiss_index_builder(hnsw_nlinks: int, pq_nblocks: int, ivf_nlists: int):
    cfg = RetrieverFaissCPU.Config(
        D,
        hnsw_nlinks=hnsw_nlinks,
        ivf_nlists=ivf_nlists,
        pq_nblocks=pq_nblocks,
        pq_nbits=nbits,
    )
    index = faiss_index_builder(cfg, D, RetrieverFaissCPU.METRICS_MAP[cfg.metric])

    if ivf_nlists > 0:
        if pq_nblocks > 0:
            assert isinstance(index, faiss.IndexIVFPQ)
        else:
            assert isinstance(index, faiss.IndexIVFFlat)

        cq = faiss.downcast_index(index.quantizer)
        if hnsw_nlinks > 0:
            assert isinstance(cq, faiss.IndexHNSWFlat)
        else:
            assert isinstance(cq, faiss.IndexFlat)
    else:
        if pq_nblocks > 0:
            if hnsw_nlinks > 0:
                assert isinstance(index, faiss.IndexHNSWPQ)
            else:
                assert isinstance(index, faiss.IndexPQ)
        else:
            if hnsw_nlinks > 0:
                assert isinstance(index, faiss.IndexHNSWFlat)
            else:
                assert isinstance(index, faiss.IndexFlat)


class TestRetrieverFaissCPU:
    @pytest.fixture
    def retriever(self):
        cfg = RetrieverFaissCPU.Config(D)
        index = faiss.IndexIDMap(faiss.IndexFlatL2(D))
        return RetrieverFaissCPU(index, cfg)

    def test__len__(self, retriever: RetrieverFaissCPU, v: np.ndarray):
        idx = np.arange(N)
        assert len(retriever) == 0
        retriever.index.add_with_ids(v, idx)
        assert len(retriever) == N
        retriever.index.add_with_ids(v, idx)
        assert len(retriever) == N * 2

    @pytest.mark.parametrize("metric", RetrieverFaissCPU.METRICS_MAP.keys())
    @pytest.mark.parametrize("opq", [False, True])
    @pytest.mark.parametrize("pca", [False, True])
    @pytest.mark.parametrize("pca_dim", [0, 4])
    def test_build(self, metric: str, opq: bool, pca: bool, pca_dim: int):
        if opq and pca:
            with pytest.raises(ValueError):
                RetrieverFaissCPU.Config(
                    D, metric=metric, opq=opq, pca=pca, pca_dim=pca_dim
                )
        else:
            cfg = RetrieverFaissCPU.Config(
                D, metric=metric, opq=opq, pca=pca, pca_dim=pca_dim
            )
            index = RetrieverFaissCPU.build(cfg).index

            if opq:
                assert isinstance(index, faiss.IndexPreTransform)
                idmap_index = faiss.downcast_index(index.index)
                assert isinstance(idmap_index, faiss.IndexIDMap)
                assert isinstance(
                    faiss.downcast_index(idmap_index.index), faiss.IndexFlat
                )
                assert isinstance(
                    faiss.downcast_VectorTransform(index.chain.at(0)), faiss.OPQMatrix
                )
            elif pca:
                assert cfg.pca_dim == D if pca_dim <= 0 else pca_dim
                assert isinstance(index, faiss.IndexPreTransform)
                idmap_index = faiss.downcast_index(index.index)
                assert isinstance(idmap_index, faiss.IndexIDMap)
                assert isinstance(
                    faiss.downcast_index(idmap_index.index), faiss.IndexFlat
                )
                assert isinstance(
                    faiss.downcast_VectorTransform(index.chain.at(0)), faiss.PCAMatrix
                )
            else:
                assert isinstance(index, faiss.IndexIDMap)
                assert isinstance(faiss.downcast_index(index.index), faiss.IndexFlat)

    @pytest.mark.parametrize("hnsw_nlinks", [0, 4])
    @pytest.mark.parametrize("ivf_nlists", [0, nlists])
    @pytest.mark.parametrize("opq", [False, True])
    def test_set_nprobe(self, hnsw_nlinks: int, ivf_nlists: int, opq: bool):
        retriever = RetrieverFaissCPU.build(
            RetrieverFaissCPU.Config(
                D, hnsw_nlinks=hnsw_nlinks, ivf_nlists=ivf_nlists, opq=opq
            )
        )
        index = retriever.index
        if ivf_nlists > 0:
            assert faiss.extract_index_ivf(index).nprobe == 1
            retriever.set_nprobe(8)
            assert faiss.extract_index_ivf(index).nprobe == 8
        else:
            # Do nothing
            retriever.set_nprobe(8)

    @pytest.mark.parametrize("hnsw_nlinks", [0, 4])
    @pytest.mark.parametrize("ivf_nlists", [0, nlists])
    @pytest.mark.parametrize("opq", [False, True])
    def test_set_efsearch(self, hnsw_nlinks: int, ivf_nlists: int, opq: bool):
        retriever = RetrieverFaissCPU.build(
            RetrieverFaissCPU.Config(
                D, hnsw_nlinks=hnsw_nlinks, ivf_nlists=ivf_nlists, opq=opq
            )
        )
        index = retriever.index
        if hnsw_nlinks > 0:
            if ivf_nlists > 0:
                hnsw = faiss.downcast_index(
                    faiss.extract_index_ivf(index).quantizer
                ).hnsw
            else:
                if opq:
                    idmap = faiss.downcast_index(index.index)
                else:
                    idmap = index
                hnsw = faiss.downcast_index(idmap.index).hnsw
            assert hnsw.efSearch == 16
            retriever.set_efsearch(64)
            assert hnsw.efSearch == 64
        else:
            # Do nothing
            retriever.set_efsearch(64)

    @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
    @pytest.mark.parametrize("dtype", [np.float32, np.float16])
    def test_normalize(self, metric: str, dtype, v: np.ndarray):
        cfg = RetrieverFaissCPU.Config(D, metric=metric)
        if metric == "l2":
            index = faiss.IndexFlatL2(D)
        else:
            index = faiss.IndexFlatIP(D)
        retriever = RetrieverFaissCPU(index, cfg)
        v = v.astype(dtype)
        nv = retriever.normalize(v)
        assert np.issubdtype(nv.dtype, np.float32)
        if metric == "cos":
            assert np.allclose((nv**2).sum(-1), np.ones(N))
        else:
            assert np.allclose(nv, v)

    @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
    def test_train(self, metric: str, v: np.ndarray):
        cfg = RetrieverFaissCPU.Config(D, metric=metric)
        if metric == "l2":
            index = faiss.IndexFlatL2(D)
        else:
            index = faiss.IndexFlatIP(D)
        retriever = RetrieverFaissCPU(index, cfg)
        retriever.train(v)
        assert retriever.index.is_trained

    @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
    @pytest.mark.parametrize("ids", [None, np.arange(N).astype(np.int64) + 1])
    def test_add(self, metric: str, ids: Optional[np.ndarray], v: np.ndarray):
        cfg = RetrieverFaissCPU.Config(D, metric=metric)
        if metric == "l2":
            index = faiss.IndexFlatL2(D)
        else:
            index = faiss.IndexFlatIP(D)
        index = faiss.IndexIDMap(index)
        retriever = RetrieverFaissCPU(index, cfg)
        retriever.add(v, ids=ids)
        dist, idxs = retriever.index.search(v, k=1)
        if metric == "l2":
            assert np.allclose(dist, np.zeros_like(dist))
            expected_ids = np.arange(N)
            if ids is not None:
                expected_ids += 1
            assert np.array_equal(idxs, expected_ids[:, None])
        else:
            dtable = (v[:, None] * v[None, :]).sum(-1)
            argmax_ids = np.argmax(dtable, axis=-1)
            expected_dist = dtable[np.arange(N), argmax_ids]
            if ids is None:
                expected_ids = argmax_ids
            else:
                expected_ids = argmax_ids + 1
            assert np.array_equal(idxs, expected_ids[:, None])
            assert np.allclose(dist, expected_dist[:, None])

    @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
    def test_search(self, metric: str, v: np.ndarray):
        cfg = RetrieverFaissCPU.Config(D, metric=metric)
        if metric == "l2":
            index = faiss.IndexFlatL2(D)
        else:
            index = faiss.IndexFlatIP(D)
        index = faiss.IndexIDMap(index)
        retriever = RetrieverFaissCPU(index, cfg)
        retriever.add(v)
        dist, ids = retriever.search(v, k=1)
        if metric == "l2":
            assert np.allclose(dist, np.zeros_like(dist))
        else:
            dtable = (v[:, None] * v[None, :]).sum(-1)
            expected_ids = np.argmax(dtable, axis=-1)
            expected_dist = dtable[np.arange(N), expected_ids]
            assert np.array_equal(ids, expected_ids[:, None])
            assert np.allclose(dist, expected_dist[:, None])

    def test_io(self, tmp_path: Path, retriever: RetrieverFaissCPU, v: np.ndarray):
        idx_path = tmp_path / "test.idx"
        cfg_path = tmp_path / "test.cfg"
        v = np.random.rand(N, D).astype(np.float32)
        retriever.add(v)
        retriever.save(idx_path, cfg_path)
        new_retriever = RetrieverFaissCPU.load(idx_path, cfg_path)
        assert len(new_retriever) == len(retriever)



import abc
from dataclasses import asdict, dataclass
from os import PathLike
from typing import Any, Optional, Tuple, Type, TypeVar

import numpy as np
import yaml

from semsis import retriever

T = TypeVar("T")


class Retriever(abc.ABC):
    """Base class of retriever classes.

    Args:
        index (Any): Index object.
        cfg (Retriever.Config): Configuration dataclass.
    """

    def __init__(self, index: Any, cfg: "Config") -> None:
        self.index = index
        self.cfg = cfg

    @dataclass
    class Config:
        """Configuration of the retriever.

        - dim (int): Size of the dimension.
        - backend (str): Backend of the search engine.
        - metric (str): Distance function.
        """

        dim: int
        backend: str = "faiss-cpu"
        metric: str = "l2"

        def save(self, path: PathLike) -> None:
            """Save the configuration.

            Args:
                path (os.PathLike): File path.
            """
            with open(path, mode="w") as f:
                yaml.dump(asdict(self), f, indent=True)

        @classmethod
        def load(cls, path: PathLike):
            """Load the configuration.

            Args:
                path (os.PathLike): File path.

            Returns:
                Retriver.Config: This configuration object.
            """
            with open(path, mode="r") as f:
                config = yaml.safe_load(f)
            return cls(**config)

    @abc.abstractmethod
    def __len__(self) -> int:
        """Return the size of the index."""

    @classmethod
    @abc.abstractmethod
    def build(cls: Type[T], cfg: "Config") -> T:
        """Build this class from the given configuration.


        Args:
            cfg (Retriever.Config): Configuration.

        Returns:
            Retriever: This class with the constucted index object.
        """

    def to_gpu_train(self) -> None:
        """Transfers the index to GPUs for training."""

    def to_gpu_add(self) -> None:
        """Transfers the index to GPUs for adding vectors."""

    def to_gpu_search(self) -> None:
        """Transfers the index to GPUs for searching."""

    def to_cpu(self) -> None:
        """Transfers the index to CPUs."""

    def set_nprobe(self, nprobe: int) -> None:
        """Set nprobe parameter for IVF-family indexes.

        Args:
            nprobe (int): Number of nearest neighbor clusters that are
              probed in search time.
        """

    def set_efsearch(self, efsearch: int) -> None:
        """Set efSearch parameter for HNSW indexes.

        Args:
            efsearch (int): The depth of exploration of the search.
        """

    @abc.abstractmethod
    def normalize(self, vectors: np.ndarray) -> np.ndarray:
        """Normalize the input vectors for a backend library and the specified metric.

        Args:
            vectors (np.ndarray): Input vectors.

        Returns:
            np.ndarray: Normalized vectors.
        """

    @abc.abstractmethod
    def train(self, vectors: np.ndarray) -> None:
        """Train the index for some approximate nearest neighbor search algorithms.

        Args:
            vectors (np.ndarray): Training vectors.
        """

    @abc.abstractmethod
    def add(self, vectors: np.ndarray, ids: Optional[np.ndarray] = None) -> None:
        """Add key vectors to the index.

        Args:
            vectors (np.ndarray): Key vectors to be added.
            ids (np.ndarray, optional): Value indices.
        """

    @abc.abstractmethod
    def search(self, querys: np.ndarray, k: int = 1) -> Tuple[np.ndarray, np.ndarray]:
        """Search the k nearest neighbor vectors of the querys.

        Args:
            querys (np.ndarray): Query vectors.
            k (int): Top-k.

        Returns:
            distances (np.ndarray): Distances between the querys and the k nearest
               neighbor vectors.
            indices (np.ndarray): Indices of the k nearest neighbor vectors.
        """

    @classmethod
    def load(cls: Type[T], index_path: PathLike, cfg_path: PathLike) -> T:
        """Load the index and its configuration.

        Args:
            index_path (os.PathLike): Index file path.
            cfg_path (os.PathLike): Configuration file path.

        Returns:
            Retriever: This class.
        """
        cfg = cls.Config.load(cfg_path)
        index = cls.load_index(index_path)
        return cls(index, cfg)

    def save(self, index_path: PathLike, cfg_path: PathLike) -> None:
        """Save the index and its configuration.

        Args:
            index_path (os.PathLike): Index file path to save.
            cfg_path (os.PathLike): Configuration file path to save.
        """
        self.cfg.save(cfg_path)
        self.save_index(index_path)

    @classmethod
    @abc.abstractmethod
    def load_index(cls, path: PathLike) -> Any:
        """Load the index.

        Args:
            path (os.PathLike): Index file path.

        Returns:
            Any: Index object.
        """

    @abc.abstractmethod
    def save_index(self, path: PathLike) -> None:
        """Save the index.

        Args:
            path (os.PathLike): Index file path to save.
        """

    @classmethod
    def get_cls(cls, name: str) -> Type["Retriever"]:
        """Return the retriever class from the given name.

        Args:
            name (str): Registered name.

        Returns:
            Type[Retriever]: Retriever class.
        """
        return retriever.get_cls(name)



from typing import List, Union

from transformers import (
    AutoTokenizer,
    BatchEncoding,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)


class Tokenizer:
    def __init__(
        self, tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast]
    ) -> None:
        self.tokenizer = tokenizer

    @classmethod
    def build(cls, model_name_or_path: str) -> "Tokenizer":
        """Build the tokenizer.

        Args:
            model_name_or_path (str): Model name or path of huggingface transformer models.

        Returns:
            Tokenizer: This class.
        """
        return cls(AutoTokenizer.from_pretrained(model_name_or_path))

    def __call__(self, *args, **kwargs):
        return self.tokenizer(*args, **kwargs)

    def tokenize(self, sentence: str) -> List[int]:
        """Tokenize and convert a sentence to the token sequence.

        Args:
            sentence (str): An input sentence.

        Returns:
            List[int]: A token ID sequence.
        """
        return self.tokenizer.encode(
            sentence, add_special_tokens=False, truncation=True
        )

    def collate(self, samples: List[List[int]]) -> BatchEncoding:
        """Make a mini-batch from samples.

        Args:
            samples (List[List[int]]): Token sequences.

        Returns:
            BatchEncoding: A mini-batch.
        """
        batch = {}
        for sample in samples:
            item = self.tokenizer.prepare_for_model(
                sample,
                None,
                add_spenical_tokens=True,
                padding=False,
                truncation=True,
                pad_to_multiple_of=None,
                return_attention_mask=False,
                return_tensors=None,
            )
            for key, value in item.items():
                if key not in batch:
                    batch[key] = []
                batch[key].append(value)

        return self.tokenizer.pad(batch, padding=True, return_tensors="pt")



import time

import pytest

from semsis.utils import Stopwatch


class TestStopwatch:
    def test__init__(self) -> None:
        timer = Stopwatch()
        assert timer._acc_time == 0.0

    def test_reset(self) -> None:
        timer = Stopwatch()
        assert timer._acc_time == 0.0
        timer._acc_time = 1.0
        assert timer._acc_time != 0.0
        timer.reset()
        assert timer._acc_time == 0.0

    def test_measure(self, monkeypatch: pytest.MonkeyPatch):
        t = 0.0

        def perf_counter_mock() -> float:
            return t

        monkeypatch.setattr(time, "perf_counter", perf_counter_mock)

        timer = Stopwatch()
        for _ in range(10):
            with timer.measure():
                t += 1.0
        assert timer.total == 10.0



import h5py
import numpy as np
import pytest
from numpy.typing import DTypeLike

from semsis.kvstore import KVStore

D = 8
N = 16


class TestKVStore:
    @pytest.fixture
    def f(self, tmp_path_factory: pytest.TempPathFactory):
        tmpdir = tmp_path_factory.mktemp("kvstore")
        return h5py.File(str(tmpdir / "tmp.hdf5"), mode="w")

    def test__init__(self, f: h5py.File):
        kvstore = KVStore(f)
        assert kvstore.f is f

    def test__len__(self, f: h5py.File):
        kvstore = KVStore(f)
        kvstore.new(D)
        assert len(kvstore) == 0

    def test_filename(self, f: h5py.File):
        kvstore = KVStore(f)
        assert kvstore.filename == f.filename

    @pytest.mark.parametrize("dtype", [np.float32, np.float16])
    def test_dtype(self, f: h5py.File, dtype):
        kvstore = KVStore(f)
        kvstore.new(D, dtype=dtype)
        assert np.issubdtype(kvstore.dtype, dtype)

    @pytest.mark.parametrize("dtype", [np.float32, np.float16])
    def test_new(self, f: h5py.File, dtype: DTypeLike):
        kvstore = KVStore(f)
        kvstore.new(D, dtype=dtype)
        k = kvstore.key
        v = kvstore.value
        assert isinstance(k, h5py.Dataset)
        assert isinstance(v, h5py.Dataset)
        assert np.issubdtype(k.dtype, dtype)
        assert np.issubdtype(v.dtype, np.int64)
        assert list(k.shape) == [0, D]
        assert list(v.shape) == [0]
        with pytest.raises(ValueError):
            kvstore.new(D)

    def test_add(self, f: h5py.File):
        keys = np.random.rand(N, D)
        values = np.ones(N).astype(np.int64)

        kvstore = KVStore(f)
        kvstore.new(D)
        kvstore.add(keys)
        assert len(kvstore) == N
        assert np.allclose(kvstore.key[:], keys)
        assert np.array_equal(kvstore.value, np.arange(N).astype(np.int64))
        kvstore.add(keys, values)
        assert len(kvstore) == N * 2
        assert np.allclose(kvstore.key[:], np.concatenate([keys, keys], axis=0))
        assert np.array_equal(
            kvstore.value[:], np.concatenate([np.arange(N), values], axis=0)
        )

    def test_open(self, tmp_path):
        path = str(tmp_path / "tmp.hdf5")
        with KVStore.open(path, mode="w") as kvstore:
            assert isinstance(kvstore.f, h5py.File)
            assert kvstore.f.name is not None
        assert kvstore.f.name is None

    def test_close(self, f: h5py.File):
        kvstore = KVStore(f)
        assert kvstore.f.name is not None
        kvstore.close()
        assert kvstore.f.name is None



import pickle
from dataclasses import asdict, dataclass
from os import PathLike
from pathlib import Path
from typing import Any, Optional, Tuple

import numpy as np
import pytest
import yaml

from semsis.retriever import load_backend_from_config, register
from semsis.retriever.base import Retriever

D = 8


class RetrieverMock(Retriever):
    def __len__(self) -> int:
        ...

    @classmethod
    def build(cls, cfg: "Retriever.Config"):
        ...

    def normalize(self, vectors: np.ndarray) -> np.ndarray:
        ...

    def train(self, vectors: np.ndarray) -> None:
        ...

    def add(self, vectors: np.ndarray, ids: Optional[np.ndarray] = None) -> None:
        ...

    def search(self, querys: np.ndarray, k: int = 1) -> Tuple[np.ndarray, np.ndarray]:
        ...

    @classmethod
    def load_index(cls, path: PathLike) -> Any:
        with open(path, mode="rb") as f:
            return pickle.load(f)

    def save_index(self, path: PathLike) -> None:
        with open(path, mode="wb") as f:
            pickle.dump(self.index, f)


class TestRetriever:
    @pytest.fixture
    def index(self):
        return object

    def test__init__(self, index):
        cfg = RetrieverMock.Config(D)
        retriever = RetrieverMock(index, cfg)
        assert retriever.index == index
        assert retriever.cfg == cfg

    class TestConfig:
        @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
        def test_save(self, tmp_path: Path, metric: str):
            cfg_path = tmp_path / "cfg.yaml"
            cfg = RetrieverMock.Config(D, metric=metric)
            cfg.save(cfg_path)
            with open(cfg_path, mode="r") as f:
                new_cfg = yaml.safe_load(f)
            assert asdict(cfg) == new_cfg

        @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
        def test_load(self, tmp_path: Path, metric: str):
            cfg_path = tmp_path / "cfg.yaml"
            cfg = RetrieverMock.Config(D, metric=metric)
            cfg.save(cfg_path)
            new_cfg = RetrieverMock.Config.load(cfg_path)
            assert cfg == new_cfg

    @pytest.mark.parametrize("metric", ["l2", "ip", "cos"])
    def test_io(self, tmp_path: Path, index, metric: str):
        idx_path = tmp_path / "test.idx"
        cfg_path = tmp_path / "test.cfg"
        cfg = RetrieverMock.Config(D, metric=metric)
        retriever = RetrieverMock(index, cfg)
        retriever.save(idx_path, cfg_path)
        new_retriever = RetrieverMock.load(idx_path, cfg_path)
        assert new_retriever.index == retriever.index
        assert new_retriever.cfg == retriever.cfg


def test_load_backend_from_config(tmp_path: Path):
    backend = "mock"

    @register(backend)
    class MockClass(RetrieverMock):
        @dataclass
        class Config(RetrieverMock.Config):
            foo: str = "foo"
            bar: int = 3

    cfg = MockClass.Config(D, backend)
    cfg_path = tmp_path / "cfg.yaml"
    cfg.save(cfg_path)
    assert issubclass(load_backend_from_config(cfg_path), MockClass)



from .encoder import SentenceEncoderAvg, SentenceEncoderCls, SentenceEncoderSbert
from .kvstore import KVStore
from .retriever import RetrieverFaissCPU, RetrieverFaissGPU

__all__ = [
    "KVStore",
    "RetrieverFaissCPU",
    "RetrieverFaissGPU",
    "SentenceEncoderCls",
    "SentenceEncoderAvg",
    "SentenceEncoderSbert",
]



import faiss
import pytest
import torch

from semsis.retriever.faiss_gpu import faiss_index_to_cpu, faiss_index_to_gpu

if (
    not hasattr(faiss, "GpuIndex")
    or not torch.cuda.is_available()
    or faiss.get_num_gpus() <= 0
):
    pytest.skip("Skipping faiss-gpu tests", allow_module_level=True)

D = 8
N = 10


@pytest.fixture
def flat_index():
    return faiss.IndexFlatL2(D)


# TODO: check the precompute and fp16 options
@pytest.mark.parametrize("num_gpus", [-1, 0, 1, 2])
def test_faiss_index_to_gpu(flat_index: faiss.Index, num_gpus: int, fp16: bool):
    gpu_index = faiss_index_to_gpu(flat_index)
    ngpus = faiss.get_num_gpus() if num_gpus == -1 else num_gpus
    if ngpus == 1:
        assert isinstance(gpu_index, faiss.GpuIndex)
        assert gpu_index.getDevice() == 0
    else:
        assert isinstance(gpu_index, faiss.IndexReplicas)
        nshards = gpu_index.count()
        for i in range(nshards):
            gpu_index_i = faiss.downcast_index(gpu_index.at(i))
            assert isinstance(gpu_index_i, faiss.GpuIndex)
            assert gpu_index_i.getDevice() == i


def test_faiss_index_to_cpu(flat_index: faiss.Index):
    gpu_index = faiss_index_to_gpu(flat_index, 1)
    assert isinstance(gpu_index, faiss.GpuIndex)
    cpu_index = faiss_index_to_cpu(gpu_index)
    assert not isinstance(cpu_index, faiss.GpuIndex)
    assert not isinstance(cpu_index, faiss.Index)


# TODO: add tests for RetrieverFaissGPU



#!/usr/bin/env python3
import contextlib
import logging
import math
import os
import sys
from argparse import ArgumentParser, Namespace

import numpy as np
import torch
from tqdm import tqdm

from semsis.kvstore import KVStore
from semsis.retriever import Retriever
from semsis.utils import Stopwatch

logging.basicConfig(
    format="| %(asctime)s | %(levelname)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level="INFO",
    stream=sys.stdout,
)
logger = logging.getLogger("semsis.cli.build_retriever")


def parse_args() -> Namespace:
    parser = ArgumentParser()
    # fmt: off
    parser.add_argument("--kvstore", metavar="FILE", nargs="+", required=True,
                        help="Path to key--value store files.")
    parser.add_argument("--index-path", metavar="FILE", required=True,
                        help="Path to an index file.")
    parser.add_argument("--trained-index-path", metavar="FILE",
                        help="Path to a trained index file. If this option is not specified, "
                        "the final index path will be set.")
    parser.add_argument("--config-path", metavar="FILE", required=True,
                        help="Path to a configuration file.")
    parser.add_argument("--cpu", action="store_true",
                        help="Only use CPU.")
    parser.add_argument("--fp16", action="store_true",
                        help="Use FP16.")
    parser.add_argument("--backend", metavar="NAME", type=str, default="faiss-cpu",
                        help="Backend of the search engine.")
    parser.add_argument("--metric", metavar="TYPE", default="l2", choices=["l2", "ip", "cos"],
                        help="Distance function.")
    parser.add_argument("--chunk-size", metavar="N", type=int, default=1000000,
                        help="The number of data to be loaded at a time.")
    parser.add_argument("--train-size", metavar="N", type=int, default=1000000,
                        help="The number of training data.")
    parser.add_argument("--hnsw-nlinks", metavar="N", type=int, default=0,
                        help="[HNSW] The number of links per node.")
    parser.add_argument("--ivf-nlists", metavar="N", type=int, default=0,
                        help="[IVF] The number of centroids")
    parser.add_argument("--pq-nblocks", metavar="N", type=int, default=0,
                        help="[PQ] The number of sub-vectors")
    parser.add_argument("--opq", action="store_true",
                        help="Use OPQ to minimize the quantization error.")
    parser.add_argument("--pca", action="store_true",
                        help="Use PCA to reduce the dimension size.")
    parser.add_argument("--pca-dim", metavar="N", type=int, default=-1,
                        help="The dimension size which is reduced by PCA.")
    parser.add_argument("--append-sequential", action="store_true",
                        help="Append entries from the tail.")
    parser.add_argument("--save-checkpoint", action="store_true",
                        help="Save checkpoint after adding each key--value set.")
    # fmt: on
    return parser.parse_args()


def train_retriever(
    args: Namespace, training_vectors: np.ndarray, use_gpu: bool = False
) -> Retriever:
    train_size, dim = training_vectors.shape
    retriever_type = Retriever.get_cls(args.backend)
    dim = training_vectors.shape[1]
    cfg_dc = retriever_type.Config
    cfg_dict = {"dim": dim, "backend": args.backend, "metric": args.metric}

    def set_if_hasattr(name: str):
        if hasattr(cfg_dc, name):
            cfg_dict[name] = getattr(args, name)

    set_if_hasattr("hnsw_nlinks")
    set_if_hasattr("ivf_nlists")
    set_if_hasattr("pq_nblocks")
    set_if_hasattr("opq")
    set_if_hasattr("pca")
    set_if_hasattr("pca_dim")
    set_if_hasattr("fp16")
    cfg = retriever_type.Config(**cfg_dict)
    logger.info(cfg)
    logger.info(f"Input vector: {cfg.dim} dimension")

    if getattr(args, "trained_index_path", None):
        trained_index_path = args.trained_index_path
    else:
        trained_index_path = args.index_path

    if os.path.exists(trained_index_path):
        trained_retriever = retriever_type.load(trained_index_path, args.config_path)
        if trained_retriever.cfg == cfg:
            logger.info(f"Load trained index: {trained_index_path}")
            return trained_retriever
        raise FileExistsError(trained_index_path)

    retriever = retriever_type.build(cfg)
    if use_gpu:
        retriever.to_gpu_train()

    timer = Stopwatch()
    logger.info(f"Train a retriever from {train_size:,} datapoints.")
    with timer.measure():
        retriever.train(training_vectors)
    logger.info(f"Training done in {timer.total:.1f} seconds")
    retriever.save(trained_index_path, args.config_path)
    return retriever_type.load(trained_index_path, args.config_path)


def main(args: Namespace) -> None:
    logger.info(args)

    timer = Stopwatch()
    use_gpu = torch.cuda.is_available() and not args.cpu
    chunk_size = args.chunk_size
    with contextlib.ExitStack() as stack:
        kvstores = [
            stack.enter_context(KVStore.open(fname, mode="r")) for fname in args.kvstore
        ]
        training_vectors = np.concatenate(
            [
                kvstore.key[: min(args.train_size // len(kvstores), len(kvstore))]
                for kvstore in kvstores
            ]
        )
        retriever = train_retriever(args, training_vectors, use_gpu=use_gpu)
        if use_gpu:
            retriever.to_gpu_add()

        logger.info(f"Build a retriever in {args.index_path}")
        for kvstore in kvstores:
            logger.info(f"Add {len(kvstore):,} vectors from {kvstore.filename}")
            offset = len(retriever) if args.append_sequential else 0
            with timer.measure():
                for i in tqdm(
                    range(math.ceil(len(kvstore) / chunk_size)),
                    desc="Building a retriever",
                ):
                    start_idx = i * chunk_size
                    end_idx = min(start_idx + chunk_size, len(kvstore))
                    num_added = end_idx - start_idx
                    logger.info(f"Add vectors: {num_added:,} / {len(kvstore):,}")
                    retriever.add(
                        kvstore.key[start_idx:end_idx],
                        kvstore.value[start_idx:end_idx] + offset,
                    )
                    logger.info(f"Retriever index size: {len(retriever):,}")

            if args.save_checkpoint:
                retriever.save(args.index_path, args.config_path)

        if not args.save_checkpoint:
            retriever.save(args.index_path, args.config_path)

        logger.info(f"Added {len(retriever):,} datapoints")
        logger.info(f"Retriever index size: {len(retriever):,}")
    logger.info(f"Built the retriever in {timer.total:.1f} seconds")


def cli_main() -> None:
    args = parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



import abc
from typing import Dict, List

import torch
import torch.nn as nn
from sentence_transformers import SentenceTransformer
from torch import Tensor
from transformers import AutoModel

from semsis import registry

from .tokenizer import Tokenizer

register, get_cls = registry.setup("sentence_encoder")


class SentenceEncoder(nn.Module, metaclass=abc.ABCMeta):
    """Huggingface model wrapper.

    Args:
        name_or_path (str): Model name or path.
    """

    def __init__(self, name_or_path: str) -> None:
        super().__init__()
        self.name_or_path = name_or_path
        self.load_model(name_or_path)
        self.freeze()

    @property
    def device(self) -> torch.device:
        """Return the device of this model."""
        return self.model.device

    def load_model(self, name_or_path: str) -> None:
        """Load the model and tokenizer.

        Args:
            name_or_path (str): Model name or path.
        """
        self.model = AutoModel.from_pretrained(name_or_path)
        self.tokenizer = Tokenizer.build(name_or_path)

    def freeze(self) -> None:
        for p in self.parameters():
            if getattr(p, "requires_grad", None) is not None:
                p.requires_grad = False
        self.eval()

    @classmethod
    def build(cls, model_name_or_path: str, representation: str) -> "SentenceEncoder":
        """Build the sentence encoder model.

        Args:
            model_name_or_path (str): Model name or path of huggingface transformer models.
            representation (str): Type of the sentence representation.

        Returns:
            SentenceEncoder: This class.
        """
        return get_cls(representation)(model_name_or_path)

    def encode(self, sentences: List[str]) -> Tensor:
        """Encode sentences into their sentence vectors.

        Args:
            sentences (List[str]): Input sentences.

        Returns:
            Tensor: Sentence vectors of shape `(batch_size, embed_dim)`.
        """
        batch = self.tokenizer(
            sentences, return_tensors="pt", padding=True, truncation=True
        ).to(self.device)
        return self.forward(batch)

    @abc.abstractmethod
    def forward(self, *args, **kwargs) -> Tensor:
        """Return the feature vectors of the given inputs.

        Returns:
            Tensor: Output tensor of shape `(batch_size, embed_dim)`.
        """

    def get_embed_dim(self) -> int:
        """Get the size of the embedding dimension.

        Returns:
            int: The size of the embedding dimension.
        """
        if hasattr(self.model.config, "hidden_size"):
            return self.model.config.hidden_size
        return self.model.get_input_embeddings().embedding_dim


@register("avg")
class SentenceEncoderAvg(SentenceEncoder):
    def forward(self, net_inputs: Dict[str, Tensor]) -> Tensor:
        """Return the feature vectors of the given inputs.

        Args:
            net_inputs (Dict[str, Tensor]): The inputs of huggingface models.

        Returns:
            Tensor: Output tensor of shape `(batch_size, embed_dim)`.
        """
        net_outputs = self.model(**net_inputs)
        non_pad_mask = net_inputs["attention_mask"]
        active_hiddens = net_outputs["last_hidden_state"] * non_pad_mask.unsqueeze(-1)
        return active_hiddens.sum(dim=1) / non_pad_mask.sum(dim=1, keepdim=True)


@register("cls")
class SentenceEncoderCls(SentenceEncoder):
    def forward(self, net_inputs: Dict[str, Tensor]) -> Tensor:
        """Return the feature vectors of the given inputs.

        Args:
            net_inputs (Dict[str, Tensor]): The inputs of huggingface models.

        Returns:
            Tensor: Output tensor of shape `(batch_size, embed_dim)`.
        """
        net_outputs = self.model(**net_inputs)
        return net_outputs["pooler_output"]


@register("sbert")
class SentenceEncoderSbert(SentenceEncoder):
    def load_model(self, name_or_path: str) -> None:
        """Load the model and tokenizer.

        Args:
            name_or_path (str): Model name or path.
        """
        self.model = SentenceTransformer(name_or_path)
        self.tokenizer = Tokenizer(self.model.tokenizer)

    def forward(self, net_inputs: Dict[str, Tensor]) -> Tensor:
        """Return the feature vectors of the given inputs.

        Args:
            net_inputs (Dict[str, Tensor]): The inputs of huggingface models.

        Returns:
            Tensor: Output tensor of shape `(batch_size, embed_dim)`.
        """
        net_outputs = self.model(net_inputs)
        return net_outputs["sentence_embedding"]

    def get_embed_dim(self) -> int:
        """Get the size of the embedding dimension.

        Returns:
            int: The size of the embedding dimension.
        """
        return self.model.get_sentence_embedding_dimension()



from os import PathLike
from typing import Type

import yaml

from semsis import registry

register, get_cls = registry.setup("retriever")

from .base import Retriever
from .faiss_cpu import RetrieverFaissCPU
from .faiss_gpu import RetrieverFaissGPU


def load_backend_from_config(cfg_path: PathLike) -> Type[Retriever]:
    """Load the backend retriever type from the configuration file.

    Args:
        cfg_path (os.PathLike): Path to the configuration file.

    Returns:
        Type[Retriever]: The backend retriever type.
    """
    with open(cfg_path, mode="r") as f:
        cfg = yaml.safe_load(f)
    return get_cls(cfg["backend"])


__all__ = [
    "Retriever",
    "RetrieverFaissCPU",
    "RetrieverFaissGPU",
    "register",
    "get_cls",
    "load_backend_from_config",
]



import pytest
import torch
from transformers import AutoTokenizer, PreTrainedTokenizer, PreTrainedTokenizerFast

from .tokenizer import Tokenizer

BERT = "bert-base-uncased"
SBERT = "sentence-transformers/all-MiniLM-L6-v2"


@pytest.fixture(scope="module")
def bert_tokenizer():
    return AutoTokenizer.from_pretrained(BERT)


class TestTokenizer:
    @pytest.mark.parametrize("name_or_path", [BERT, SBERT])
    def test_build(self, name_or_path: str):
        tokenizer = Tokenizer.build(name_or_path)
        assert isinstance(
            tokenizer.tokenizer, (PreTrainedTokenizer, PreTrainedTokenizerFast)
        )

    def test_tokenize(self, bert_tokenizer: PreTrainedTokenizer):
        tokenizer = Tokenizer(bert_tokenizer)
        examples = ["i like apples .", "this is my pen .", "i like apples ."]
        sequences = [tokenizer.tokenize(sentence) for sentence in examples]
        assert len(sequences) == len(examples)
        assert [
            tokenizer.tokenizer.convert_ids_to_tokens(seq) for seq in sequences
        ] == [sentence.split() for sentence in examples]

    def test_collate(self, bert_tokenizer: PreTrainedTokenizer):
        tokenizer = Tokenizer(bert_tokenizer)
        examples = ["i like apples .", "this is my pen .", "i like apples ."]
        sequences = [tokenizer.tokenize(sentence) for sentence in examples]
        batch = tokenizer.collate(sequences)
        expected = tokenizer.tokenizer(
            examples, return_tensors="pt", padding=True, truncation=True
        )
        assert batch.keys() == expected.keys()
        assert [
            torch.equal(v, expected[k]) if torch.is_tensor(v) else v == expected[k]
            for k, v in batch.items()
        ]



import logging
from dataclasses import dataclass
from typing import Optional, Union

import faiss
import numpy as np
import torch

from semsis.retriever import register
from semsis.retriever.faiss_cpu import RetrieverFaissCPU

logger = logging.getLogger(__name__)

GpuIndex = (
    faiss.GpuIndex if hasattr(faiss, "GpuIndex") else faiss.Index
)  # avoid error on faiss-cpu
ShardedGpuIndex = Union[GpuIndex, faiss.IndexReplicas]


def faiss_index_to_gpu(
    index: faiss.Index, num_gpus: int = -1, precompute: bool = False, fp16: bool = False
) -> ShardedGpuIndex:
    """Transfers the index from CPU to GPU.

    Args:
        index (faiss.Index): Faiss index.
        num_gpus (int): Number of GPUs to use. `-1` uses all devices.
        precompute (bool): Uses the precompute table for IVF-family.
        fp16 (bool): Use fp16.

    Returns:
        faiss.GpuIndex | faiss.IndexReplicas: Faiss index.
    """
    co = faiss.GpuMultipleClonerOptions()
    co.useFloat16 = fp16
    co.useFloat16CoarseQuantizer = fp16
    co.indicesOptions = faiss.INDICES_CPU
    co.shard = True
    if precompute:
        co.usePrecomputed = precompute

    return faiss.index_cpu_to_all_gpus(index, co, ngpu=num_gpus)


def faiss_index_to_cpu(index: ShardedGpuIndex) -> faiss.Index:
    """Transfers the index from GPU to CPU.

    Args:
        index (faiss.GpuIndex | faiss.IndexReplicas): faiss index.

    Returns:
        faiss.Index: faiss index.
    """
    return faiss.index_gpu_to_cpu(index)


@register("faiss-gpu")
class RetrieverFaissGPU(RetrieverFaissCPU):
    """Faiss GPU retriever class.

    This class extend the faiss behavior for efficiency.

    Args:
        index (faiss.Index): Index object.
        cfg (RetrieverFaissGPU.Config): Configuration dataclass.
    """

    cfg: "RetrieverFaissGPU.Config"

    def __init__(self, index: faiss.Index, cfg: "RetrieverFaissGPU.Config") -> None:
        super().__init__(index, cfg)
        self.A: Optional[torch.Tensor] = None
        self.b: Optional[torch.Tensor] = None
        self.gpu_ivf_index: Optional[ShardedGpuIndex] = None

    @dataclass
    class Config(RetrieverFaissCPU.Config):
        """Configuration of the retriever.

        - dim (int): Size of the dimension.
        - backend (str): Backend of the search engine.
        - metric (str): Distance function.
        - hnsw_nlinks (int): [HNSW] Number of links for each node.
            If this value is greater than 0, HNSW will be used.
        - ivf_nlists (int): [IVF] Number of centroids.
        - pq_nblocks (int): [PQ] Number of sub-vectors to be splitted.
        - pq_nbits (int): [PQ] Size of codebooks for each sub-space.
            Usually 8 bit is employed; thus, each codebook has 256 codes.
        - opq (bool): [OPQ] Use OPQ pre-transformation which minimizes the quantization error.
        - pca (bool): [PCA] Use PCA dimension reduction.
        - pca_dim (int): [PCA] Dimension size which is reduced by PCA.
        - fp16 (bool): Use FP16.
        """

        fp16: bool = False

    def to_gpu_train(self) -> None:
        """Transfers the faiss index to GPUs for training.

        This speeds up training of IVF clustering and the PQ codebook:

        * IVF learns the nlists centroids by k-means.

        * PQ can be used for two components: OPQ training and vector quantization.
          - OPQ training: OPQ rotates the input vectors as a pre-transform and the
              rotation matrix is trained to minimize the PQ reconstruction error. If the
              ProductQuantizer object is given in OPQ training, it is finally set
              `ProductQuantizer::Train_hot_start` flag that initializes the codebook by
              the state in the last iteration of OPQ training.
          - Vector Quantization: It learns the PQ codes that is actually added to the
              index. Note that IVFPQ learns codes from residual vectors from each
              centroid, which are different from OPQ.
        """
        if self.cfg.ivf:
            # IVF* learns the nlists centroids from d-dimensional vectors.
            # clustering_index passignments by computing distance between `nlists x d`
            # and `ntrain x d` that would be large, so it is sharded.
            ivf = faiss.extract_index_ivf(self.index)
            clustering_index = faiss_index_to_gpu(
                faiss.IndexFlat(ivf.d, self.METRICS_MAP[self.cfg.metric]),
                fp16=self.cfg.fp16,
            )
            ivf.clustering_index = clustering_index

        if self.cfg.pq:
            index = (
                self.index.index
                if isinstance(self.index, faiss.IndexPreTransform)
                else self.index
            )
            pq: faiss.ProductQuantizer = faiss.downcast_index(index).pq

            # PQ splits input vectors into dsub-dimensional sub-vectors and assigns
            # quantization codes in each sub-space. In addition, PQ is trained from
            # sampled vectors and all training vectors are not used. Therefore, a single
            # GPU is used for PQ assignment because the GPU memory footprint is small
            # (`ksub x dsub`, where ksub is the codebook size and typically =256).
            pq.assign_index = faiss_index_to_gpu(
                faiss.IndexFlatL2(pq.dsub), num_gpus=1, fp16=self.cfg.fp16
            )
            if self.cfg.opq:
                opq: faiss.OPQMatrix = faiss.downcast_VectorTransform(
                    self.index.chain.at(0)
                )
                if self.cfg.ivf:
                    # IVFPQ learns PQ from the different vectors, so this PQ is not
                    # shared with IVFPQ.
                    opq_pq = faiss.ProductQuantizer(opq.d_out, opq.M, pq.nbits)
                    opq_pq.assign_index = faiss_index_to_gpu(
                        faiss.IndexFlatL2(opq_pq.dsub), num_gpus=1, fp16=self.cfg.fp16
                    )
                else:
                    # Otherwise, PQ codes are initialized by the last state of OPQ
                    # training and PQ training starts from that initialized codebook.
                    opq_pq = pq
                opq.pq = opq_pq

    def to_gpu_add(self) -> None:
        """Transfers the faiss index to GPUs for adding."""

        if self.cfg.transform:
            vt: faiss.LinearTransform = faiss.downcast_VectorTransform(
                faiss.downcast_index(self.index).chain.at(0)
            )
            self.A = torch.from_numpy(
                faiss.vector_to_array(vt.A).reshape(vt.d_out, vt.d_in).T
            )
            self.b = torch.from_numpy(faiss.vector_to_array(vt.b))

            self.A = self.A.cuda()
            self.b = self.b.cuda()
            if self.cfg.fp16:
                self.A = self.A.half()
                self.b = self.b.half()

        if self.cfg.ivf:
            ivf = faiss.extract_index_ivf(self.index)
            if self.cfg.hnsw:
                # Temporarily replace the HNSW coarse quantizer with the flat index.
                hnsw_cq = faiss.downcast_index(ivf.quantizer)
                ivf.quantizer = faiss.downcast_index(hnsw_cq.storage)
                self.gpu_ivf_index = faiss_index_to_gpu(ivf, fp16=self.cfg.fp16)

                # After transfering the IVF index with the flat coarse quantizer,
                # the coarse quantizer of the master IVF index is restored to HNSW.
                ivf.quantizer = hnsw_cq
            else:
                self.gpu_ivf_index = faiss_index_to_gpu(ivf, fp16=self.cfg.fp16)
            self.gpu_ivf_index.reset()

    def to_gpu_search(self) -> None:
        """Transfers the faiss index to GPUs for searching."""
        if self.cfg.ivf and self.cfg.hnsw:
            ivf_index: faiss.IndexIVF = faiss.extract_index_ivf(self.index)
            ivf_index.quantizer = faiss.downcast_index(
                faiss.downcast_index(ivf_index.quantizer).storage
            )
        self.index = faiss_index_to_gpu(self.index, fp16=self.cfg.fp16)
        logger.info(f"The retriever index is on the GPU.")

    def to_cpu(self) -> None:
        """Transfers the faiss index to CPUs."""
        self.index = faiss_index_to_cpu(self.index)

    def rotate(self, x: torch.Tensor, shard_size: int = 2**20) -> torch.Tensor:
        """Rotate the input vectors instead of faiss.IndexPreTransform.

        Args:
            x (torch.Tensor): Input vectors of shpae `(n, D)`
            shard_size (int): Number of rotated vectors at once.
              The default size is 2**20 (Each shard would take 2 GiB when D=256).

        Returns:
            torch.Tensor: Pre-transformed vectors of shape `(n, D)`.
        """
        if self.A is None:
            return x

        x = x.type(self.A.dtype)
        x_device = x.device
        A_device = self.A.device

        # Compute rotation of `x[i:j]` while `i < n`.
        results = []
        n = x.size(0)
        if shard_size <= 0:
            shard_size = n
        i = 0
        while i < n:
            j = min(i + shard_size, n)
            xs = x[i:j]
            xs = xs.to(A_device)
            xs @= self.A
            if self.b is not None and self.b.numel() > 0:
                xs += self.b
            results.append(xs.to(x_device))
            i = j
        return torch.cat(results, dim=0)

    def add_gpu_ivf_index(self, vectors: np.ndarray, ids: np.ndarray) -> None:
        """Adds vectors to the index with the full GPU IVF index.

        This method runs as follows:
        1. Transfers a trained IVF index to GPU devices. The index has only centroid
          information and does not have any vectors.
        2. Adds the input vectors to the temporary GPU index.
        3. Copies a replica of the GPU index to CPU.
        4. Copies the added vectors with their IVF lists from 3. to the real index.
        5. Empties the storage of the GPU index. Here, the GPU index has only centroids.

        Args:
            vectors (ndarray): Key vectors to be added.
            ids (np.ndarray): Value indices.
        """
        ivf: faiss.IndexIVF = faiss.extract_index_ivf(self.index)
        self.gpu_ivf_index.add_with_ids(vectors, ids)
        cpu_ivf_index: faiss.IndexIVF = faiss_index_to_cpu(self.gpu_ivf_index)
        assert cpu_ivf_index.ntotal == vectors.shape[0]
        ivf.merge_from(cpu_ivf_index, 0)
        self.gpu_ivf_index.reset()

    def add(self, vectors: np.ndarray, ids: Optional[np.ndarray] = None) -> None:
        """Add key vectors to the index.

        Args:
            vectors (np.ndarray): Key vectors to be added.
            ids (np.ndarray, optional): Value indices.
        """
        vectors = self.normalize(vectors)
        if self.cfg.transform:
            vectors = self.rotate(torch.from_numpy(vectors)).numpy().astype(np.float32)

        if ids is None:
            ids = np.arange(len(self), len(self) + len(vectors))

        index = self.index
        if self.cfg.transform:
            index = faiss.downcast_index(index.index)

        if self.cfg.ivf and self.gpu_ivf_index is not None:
            self.add_gpu_ivf_index(vectors, ids)
        else:
            index.add_with_ids(vectors, ids)

        if self.cfg.transform:
            self.index.ntotal = index.ntotal



#!/usr/bin/env python3
import concurrent.futures
import ctypes.util
import logging
import math
import signal
import sys
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from os import PathLike
from typing import Generator, List

import numpy as np
import torch
from tqdm import tqdm
from transformers import BatchEncoding

from semsis.encoder import SentenceEncoder
from semsis.encoder.tokenizer import Tokenizer
from semsis.kvstore import KVStore
from semsis.registry import get_registry
from semsis.utils import Stopwatch

logging.basicConfig(
    format="| %(asctime)s | %(levelname)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level="INFO",
    stream=sys.stdout,
)
logger = logging.getLogger("semsis.cli.store_kv")


@dataclass
class Batch:
    """Mini-batch class.

    inputs (BatchEncoding): Model inputs.
    ids (np.ndarray): Sample IDs.
    """

    inputs: BatchEncoding
    ids: np.ndarray


@dataclass
class Dataset:
    """Dataset class.

    sequences (List[List[int]]): Token ID sequences.
    lengths (np.ndarray): Lengths of each sequence.
    """

    sequences: List[List[int]]
    lengths: np.ndarray

    def __len__(self):
        return len(self.lengths)

    def yield_batches(
        self, tokenizer: Tokenizer, batch_size: int
    ) -> Generator[Batch, None, None]:
        """Yields mini-batches.

        Args:
            tokenizer (Tokenizer): A tokenizer.
            batch_size (int): Batch size.

        Yields:
            Batch: A mini-batch.
        """
        sort_order = self.lengths.argsort()[::-1]
        num_data = len(self)
        for n in tqdm(range(math.ceil(len(self) / batch_size))):
            b, e = n * batch_size, min((n + 1) * batch_size, num_data)
            indices = sort_order[b:e]
            batch = tokenizer.collate([self.sequences[i] for i in indices])
            yield Batch(batch, indices)


def set_pdeathsig(sig: int) -> None:
    """Sets pdeathsig by prctl."""
    libc_path = ctypes.util.find_library("c")
    if libc_path is not None:
        libc = ctypes.cdll.LoadLibrary(libc_path)
        if hasattr(libc, "prctl"):
            PR_SET_PDEATHSIG = 1
            libc.prctl(PR_SET_PDEATHSIG, sig)


def prepare_dataset(
    file: PathLike,
    tokenizer: Tokenizer,
    num_workers: int = 1,
    chunk_size: int = 100000,
) -> Dataset:
    """Prepare a dataset.

    Args:
        file (os.PathLike): Input file.
        tokenizer (Tokenizer): Tokenizer.
        num_workers (int, optional): Number of workers.
        chunk_size (int): Size of data processed by each process at a time.

    Returns:
        Dataset: A dataset dataclass.
    """
    with open(file, mode="r") as f:
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=num_workers,
            initializer=set_pdeathsig,
            initargs=(signal.SIGINT,),
        ) as executor:
            sequences = list(
                tqdm(
                    executor.map(tokenizer.tokenize, f, chunksize=chunk_size),
                    desc="Preprocess the data",
                    mininterval=1,
                )
            )

    return Dataset(sequences, np.array([len(seq) for seq in sequences]))


def parse_args() -> Namespace:
    """Parses the command line arguments.

    Returns:
        Namespace: Command line arguments.
    """
    parser = ArgumentParser()
    # fmt: off
    parser.add_argument("--input", type=str, required=True,
                        help="Input file.")
    parser.add_argument("--output", type=str, default="kv.bin",
                        help="Path to the key--value store.")
    parser.add_argument("--model", type=str, default="sentence-transformers/LaBSE",
                        help="Model name")
    parser.add_argument("--representation", type=str, default="sbert",
                        choices=get_registry("sentence_encoder").keys(),
                        help="Sentence representation type.")
    parser.add_argument("--batch-size", type=int, default=128,
                        help="Batch size.")
    parser.add_argument("--fp16", action="store_true",
                        help="Use FP16.")
    parser.add_argument("--workers", type=int, default=16,
                        help="Number of workers to preprocess the data.")
    parser.add_argument("--chunk-size", type=int, default=100000,
                        help="Chunk size for multi-processing.")
    # fmt: on
    return parser.parse_args()


def main(args: Namespace) -> None:
    logger.info(args)

    timer = Stopwatch()
    encoder = SentenceEncoder.build(args.model, args.representation)
    tokenizer = encoder.tokenizer

    logger.info(f"Start preprocessing the data")
    with timer.measure():
        dataset = prepare_dataset(args.input, tokenizer, args.workers, args.chunk_size)
    logger.info(f"Dataset size: {len(dataset):,}")
    logger.info(f"Preprocessed the data in {timer.total:.1f} seconds.")

    if torch.cuda.is_available():
        encoder = encoder.cuda()
        if args.fp16:
            encoder = encoder.half()

    logger.info(f"Start storing the keys and values.")
    with KVStore.open(args.output, mode="w") as kvstore:
        kvstore.new(encoder.get_embed_dim(), np.float16 if args.fp16 else np.float32)

        timer.reset()
        with timer.measure():
            for batch in dataset.yield_batches(tokenizer, args.batch_size):
                sentence_vectors = (
                    encoder(batch.inputs.to(encoder.device)).cpu().numpy()
                )
                kvstore.add(sentence_vectors, batch.ids)
    logger.info(f"Stored the keys and values in {timer.total:.1f} seconds.")


def cli_main() -> None:
    args = parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



import contextlib
from os import PathLike
from typing import Any, Generator, Optional

import h5py
import numpy as np
from numpy.typing import DTypeLike


class KVStore:
    """Key--value store class.

    Args:
        f (h5py.File): HDF5 file object.
    """

    def __init__(self, f: h5py.File) -> None:
        self.f = f

    @property
    def filename(self) -> str:
        """Returns the file name on the disk."""
        return self.f.filename

    @property
    def key(self) -> h5py.Dataset:
        """Return the key array."""
        return self.f["key"]

    @property
    def value(self) -> h5py.Dataset:
        """Return the value array."""
        return self.f["value"]

    def __len__(self) -> int:
        """Return the length of the key array."""
        return self.key.shape[0]

    @property
    def dtype(self) -> DTypeLike:
        """Return the dtype of the key array."""
        return self.key.dtype

    def add(self, keys: np.ndarray, values: Optional[np.ndarray] = None) -> None:
        """Add the given key vectors into the key array.

        Args:
            keys (np.ndarray): The key vectors of shape `(n, dim)`.
            values (np.ndarray, optional): The value IDs of shape `(n, dim)`.
              If values are not given, value IDs will be assigned incrementally.
        """
        ntotal = len(self)
        n = len(keys)
        if values is None:
            values = np.arange(ntotal, ntotal + n)
        self.key.resize(ntotal + n, axis=0)
        self.value.resize(ntotal + n, axis=0)
        self.key[ntotal:] = keys
        self.value[ntotal:] = values

    def new(self, dim: int, dtype: DTypeLike = np.float32) -> None:
        """Create the new arrays.

        Args:
            dim (int): The dimension size.
            dtype (DtypeLike): Dtype of the key array.
        """
        self.f.create_dataset("key", shape=(0, dim), dtype=dtype, maxshape=(None, dim))
        self.f.create_dataset("value", shape=(0,), dtype=np.int64, maxshape=(None,))

    @classmethod
    @contextlib.contextmanager
    def open(cls, path: PathLike, mode: str = "r") -> Generator["KVStore", Any, None]:
        """Open a binary file of this kvstore.

        Args:
            path (os.PathLike): A path to the file.
            mode (str): Mode of this file objects.
              See https://docs.h5py.org/en/stable/high/file.html

        Yields:
            KVStore: This class.
        """
        f = h5py.File(path, mode=mode)
        self = cls(f)
        try:
            yield self
        finally:
            self.close()

    def close(self) -> None:
        """Close the file stream."""
        self.f.flush()
        self.f.close()



import contextlib
import time


class Stopwatch:
    """Stopwatch class to measure the elapsed time.

    Example:
        >>> timer = Stopwatch()
        >>> for i in range(10):
                with timer.measure():
                    time.sleep(1)
        >>> print(f"{timer.total:.3f}")
        10.000
    """

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Reset the stopwatch."""
        self._acc_time = 0.0

    @contextlib.contextmanager
    def measure(self):
        """Start to measure the time."""
        start = time.perf_counter()
        try:
            yield
        finally:
            self._acc_time += time.perf_counter() - start

    @property
    def total(self) -> float:
        """Return the total time."""
        return self._acc_time



from dataclasses import dataclass
from os import PathLike
from typing import Any, Dict, Optional, Tuple

import faiss
import numpy as np

from semsis.retriever import Retriever, register

MetricType = int


def faiss_index_builder(
    cfg: "RetrieverFaissCPU.Config", dim: int, metric: MetricType
) -> faiss.Index:
    """Build a faiss index from the given configuration.

    Args:
        cfg (RetrieverFaissCPU.Config):
        dim (int):
        metric: MetricType

    Returns:
        faiss.Index: Faiss index.
    """
    if cfg.ivf:
        if cfg.hnsw:
            cq = faiss.IndexHNSWFlat(dim, cfg.hnsw_nlinks, metric)
        else:
            cq = faiss.IndexFlat(dim, metric)

        if cfg.pq:
            return faiss.IndexIVFPQ(
                cq, dim, cfg.ivf_nlists, cfg.pq_nblocks, cfg.pq_nbits, metric
            )
        else:
            return faiss.IndexIVFFlat(cq, dim, cfg.ivf_nlists, metric)
    else:
        if cfg.pq:
            if cfg.hnsw:
                return faiss.IndexHNSWPQ(dim, cfg.pq_nblocks, cfg.hnsw_nlinks)
            return faiss.IndexPQ(dim, cfg.pq_nblocks, cfg.pq_nbits, metric)
        else:
            if cfg.hnsw:
                return faiss.IndexHNSWFlat(dim, cfg.hnsw_nlinks, metric)
            return faiss.IndexFlat(dim, metric)


@register("faiss-cpu")
class RetrieverFaissCPU(Retriever):
    """Faiss CPU retriever class.

    Args:
        index (faiss.Index): Index object.
        cfg (RetrieverFaissCPU.Config): Configuration dataclass.
    """

    index: faiss.Index
    cfg: "Config"

    METRICS_MAP: Dict[str, MetricType] = {
        "l2": faiss.METRIC_L2,
        "ip": faiss.METRIC_INNER_PRODUCT,
        "cos": faiss.METRIC_INNER_PRODUCT,
    }

    @dataclass
    class Config(Retriever.Config):
        """Configuration of the retriever.

        - dim (int): Size of the dimension.
        - backend (str): Backend of the search engine.
        - metric (str): Distance function.
        - hnsw_nlinks (int): [HNSW] Number of links for each node.
            If this value is greater than 0, HNSW will be used.
        - ivf_nlists (int): [IVF] Number of centroids.
        - pq_nblocks (int): [PQ] Number of sub-vectors to be splitted.
        - pq_nbits (int): [PQ] Size of codebooks for each sub-space.
            Usually 8 bit is employed; thus, each codebook has 256 codes.
        - opq (bool): [OPQ] Use OPQ pre-transformation which minimizes the quantization error.
        - pca (bool): [PCA] Use PCA dimension reduction.
        - pca_dim (int): [PCA] Dimension size which is reduced by PCA.
        """

        hnsw_nlinks: int = 0
        ivf_nlists: int = 0
        pq_nblocks: int = 0
        pq_nbits: int = 8
        opq: bool = False
        pca: bool = False
        pca_dim: int = 0

        def __post_init__(self):
            self.hnsw = self.hnsw_nlinks > 0
            self.ivf = self.ivf_nlists > 0
            self.pq = self.pq_nblocks > 0
            if self.opq and self.pca:
                raise ValueError("`opq` and `pca` cannot be set True at the same time.")
            self.transform = self.opq or self.pca

            if self.pca and self.pca_dim <= 0:
                self.pca_dim = self.dim

    def __len__(self) -> int:
        """Return the size of the index."""
        return self.index.ntotal

    @classmethod
    def build(cls, cfg: "Config") -> "RetrieverFaissCPU":
        """Build this class from the given configuration.

        Args:
            cfg (Retriever.Config): Configuration.

        Returns:
            RetrieverFaissCPU: This class with the constucted index object.
        """
        metric = cls.METRICS_MAP[cfg.metric]
        dim = cfg.pca_dim if cfg.pca else cfg.dim

        index = faiss_index_builder(cfg, dim, metric)
        if not cfg.ivf:
            index = faiss.IndexIDMap(index)

        if cfg.opq:
            vtrans = faiss.OPQMatrix(cfg.dim, M=cfg.pq_nblocks)
            index = faiss.IndexPreTransform(vtrans, index)
        elif cfg.pca:
            vtrans = faiss.PCAMatrix(cfg.dim, d_out=cfg.pca_dim)
            index = faiss.IndexPreTransform(vtrans, index)
        return cls(index, cfg)

    def set_nprobe(self, nprobe: int) -> None:
        """Set nprobe parameter for IVF-family indexes.

        Args:
            nprobe (int): Number of nearest neighbor clusters that are
              probed in search time.
        """
        if self.cfg.ivf:
            faiss.extract_index_ivf(self.index).nprobe = nprobe

    def set_efsearch(self, efsearch: int) -> None:
        """Set efSearch parameter for HNSW indexes.

        Args:
            efsearch (int): The depth of exploration of the search.
        """
        if self.cfg.hnsw:
            faiss.ParameterSpace().set_index_parameter(self.index, "efSearch", efsearch)

    def normalize(self, vectors: np.ndarray) -> np.ndarray:
        """Normalize the input vectors for a backend library and the specified metric.

        Args:
            vectors (np.ndarray): Input vectors.

        Returns:
            np.ndarray: Normalized vectors.
        """
        if not np.issubdtype(vectors.dtype, np.float32):
            vectors = np.array(vectors, dtype=np.float32)
        if self.cfg.metric == "cos":
            vectors /= np.fmax(np.linalg.norm(vectors, axis=-1, keepdims=True), 1e-9)
        return vectors

    def train(self, vectors: np.ndarray) -> None:
        """Train the index for some approximate nearest neighbor search algorithms.

        Args:
            vectors (np.ndarray): Training vectors.
        """
        vectors = self.normalize(vectors)
        self.index.train(vectors)

    def add(self, vectors: np.ndarray, ids: Optional[np.ndarray] = None) -> None:
        """Add key vectors to the index.

        Args:
            vectors (np.ndarray): Key vectors to be added.
            ids (np.ndarray, optional): Value indices.
        """
        vectors = self.normalize(vectors)
        if ids is None:
            ids = np.arange(len(self), len(self) + len(vectors))
        return self.index.add_with_ids(vectors, ids)

    def search(self, querys: np.ndarray, k: int = 1) -> Tuple[np.ndarray, np.ndarray]:
        """Search the k nearest neighbor vectors of the querys.

        Args:
            querys (np.ndarray): Query vectors.
            k (int): Top-k.

        Returns:
            distances (np.ndarray): Distances between the querys and the k nearest
               neighbor vectors.
            indices (np.ndarray): Indices of the k nearest neighbor vectors.
        """
        querys = self.normalize(querys)
        return self.index.search(querys, k=k)

    @classmethod
    def load_index(cls, path: PathLike) -> Any:
        """Load the index.

        Args:
            path (os.PathLike): Index file path.

        Returns:
            faiss.Index: Index object.
        """
        return faiss.read_index(str(path))

    def save_index(self, path: PathLike) -> None:
        """Saves the index.

        Args:
            path (os.PathLike): Index file path to save.
        """
        return faiss.write_index(self.index, str(path))



from .sentence_encoder import (
    SentenceEncoder,
    SentenceEncoderAvg,
    SentenceEncoderCls,
    SentenceEncoderSbert,
    get_cls,
)

__all__ = [
    "SentenceEncoder",
    "SentenceEncoderAvg",
    "SentenceEncoderCls",
    "SentenceEncoderSbert",
    "get_cls",
]



#!/usr/bin/env python3
import fileinput
import logging
import sys
from argparse import ArgumentParser, Namespace
from collections import defaultdict
from os import PathLike
from typing import Generator, List

import torch

from semsis.encoder import SentenceEncoder
from semsis.registry import get_registry
from semsis.retriever import load_backend_from_config
from semsis.utils import Stopwatch

logging.basicConfig(
    format="| %(asctime)s | %(levelname)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level="INFO",
    stream=sys.stdout,
)
logger = logging.getLogger("semsis.cli.query_interactive")


def buffer_lines(
    input: PathLike, buffer_size: int = 1
) -> Generator[List[str], None, None]:
    buf: List[str] = []
    with fileinput.input(
        [input], mode="r", openhook=fileinput.hook_encoded("utf-8")
    ) as f:
        for line in f:
            buf.append(line.strip())
            if len(buf) >= buffer_size:
                yield buf
                buf = []
        if len(buf) > 0:
            yield buf


def parse_args() -> Namespace:
    """Parses the command line arguments.

    Returns:
        Namespace: Command line arguments.
    """
    parser = ArgumentParser()
    # fmt: off
    parser.add_argument("--input", type=str, default="-",
                        help="Input file.")
    parser.add_argument("--index-path", metavar="FILE",
                        help="Path to an index file.")
    parser.add_argument("--config-path", metavar="FILE", required=True,
                        help="Path to a configuration file.")
    parser.add_argument("--model", type=str, default="sentence-transformers/LaBSE",
                        help="Model name")
    parser.add_argument("--representation", type=str, default="sbert",
                        choices=get_registry("sentence_encoder").keys(),
                        help="Sentence representation type.")
    parser.add_argument("--gpu-encode", action="store_true",
                        help="Transfer the encoder to GPUs.")
    parser.add_argument("--gpu-retrieve", action="store_true",
                        help="Transfer the retriever to GPUs.")
    parser.add_argument("--fp16", action="store_true",
                        help="Use FP16.")
    parser.add_argument("--ntrials", type=int, default=1,
                        help="Number of trials to measure the search time.")
    parser.add_argument("--topk", type=int, default=1,
                        help="Search the top-k nearest neighbor.")
    parser.add_argument("--buffer-size", type=int, default=1,
                        help="Buffer size to query at a time.")
    parser.add_argument("--msec", action="store_true",
                        help="Show the search time in milli seconds instead of seconds.")
    parser.add_argument("--efsearch", type=int, default=16,
                        help="Set the efSearch parameter for the HNSW indexes. "
                        "This corresponds to the beam width at the search time.")
    parser.add_argument("--nprobe", type=int, default=8,
                        help="Set the nprobe parameter for the IVF indexes. "
                        "This corresponds to the number of neighboring clusters to be searched.")
    # fmt: on
    return parser.parse_args()


def main(args: Namespace) -> None:
    logger.info(args)

    encoder = SentenceEncoder.build(args.model, args.representation)
    if torch.cuda.is_available() and args.gpu_encode:
        encoder = encoder.cuda()
        if args.fp16:
            encoder = encoder.half()
        logger.info(f"The encoder is on the GPU.")

    retriever_type = load_backend_from_config(args.config_path)
    retriever = retriever_type.load(args.index_path, args.config_path)
    retriever.set_efsearch(args.efsearch)
    retriever.set_nprobe(args.nprobe)
    if torch.cuda.is_available() and args.gpu_retrieve:
        retriever.to_gpu_search()
    logger.info(f"Retriever configuration: {retriever.cfg}")
    logger.info(f"Retriever index size: {len(retriever):,}")

    timers = defaultdict(Stopwatch)
    ntrials = args.ntrials
    start_id = 0
    nqueryed = 0
    acctimes = defaultdict(float)
    unit = "msec" if args.msec else "sec"
    for lines in buffer_lines(args.input, buffer_size=args.buffer_size):
        timers["encode"].reset()
        timers["retrieve"].reset()
        for _ in range(ntrials):
            with timers["encode"].measure():
                querys = encoder.encode(lines).cpu().numpy()
            with timers["retrieve"].measure():
                dists, idxs = retriever.search(querys, k=args.topk)

        for i, line in enumerate(lines):
            uid = start_id + i
            dist_str = " ".join([f"{x:.3f}" for x in dists[i].tolist()])
            idx_str = " ".join([str(x) for x in idxs[i].tolist()])
            print(f"Q-{uid}\t{line}")
            print(f"D-{uid}\t{dist_str}")
            print(f"I-{uid}\t{idx_str}")

        times = {k: timer.total for k, timer in timers.items()}
        times["search"] = sum([timer.total for timer in timers.values()])
        if args.msec:
            times = {k: t * 1000 for k, t in times.items()}
        for name, t in times.items():
            acctimes[name] += t

        nqueryed = start_id + len(lines)
        for name, c in [("encode", "E"), ("retrieve", "R"), ("search", "S")]:
            t = times[name]
            at = times[name] / ntrials
            print(f"T{c}-{start_id}:{nqueryed}\t{t:.1f} {unit}")
            print(f"AT{c}-{start_id}:{nqueryed}\t{at:.1f} {unit}")

        start_id = nqueryed

    batch_avgtimes = defaultdict(float)
    single_avgtimes = defaultdict(float)
    for k, v in acctimes.items():
        batch_avgtimes[k] = v / ntrials
        single_avgtimes[k] = v / (nqueryed * ntrials)

    for k in acctimes.keys():
        print(f"Total {k} time: {acctimes[k]:.1f} {unit}")
        print(f"Average {k} time per batch: {batch_avgtimes[k]:.1f} {unit}")
        print(f"Average {k} time per single query: {single_avgtimes[k]:.1f} {unit}")


def cli_main() -> None:
    args = parse_args()
    main(args)


if __name__ == "__main__":
    cli_main()



from typing import Any, Callable, Dict, Type, TypeVar

T = TypeVar("T")

REGISTRIES = {}


def setup(registry_name: str):
    """Setup a registry.

    Args:
        registry_name (str): Registry name for grouping classes.
    """
    REGISTRY = {}
    REGISTRIES[registry_name] = REGISTRY

    def register(name: str) -> Callable[[Type[T]], Type[T]]:
        """Register a class as the given name.

        Args:
            name (str): The name of a class.
        """

        def _register(cls: Type[T]):
            if name in REGISTRY:
                raise ValueError(
                    f"{name} already registered as {REGISTRY[name].__name__}. ({cls.__name__})"
                )
            REGISTRY[name] = cls
            return cls

        return _register

    def get_cls(name: str):
        if name not in REGISTRY:
            raise NotImplementedError(
                f"`{name}` is not registered in `{registry_name}`."
            )
        return REGISTRY[name]

    return register, get_cls


def get_registry(registry_name: str) -> Dict[str, Type[Any]]:
    """Get registry of the given name.

    Args:
        registry_name (str): Registry name to be returned.

    Returns:
        Dict[str, Type[Any]]: Class mapper from registered name to its corresponding class.
    """
    return REGISTRIES[registry_name]



from typing import Dict

import pytest
import torch
from torch import Tensor
from transformers import BertModel, BertTokenizerFast

from .sentence_encoder import (
    SentenceEncoder,
    SentenceEncoderAvg,
    SentenceEncoderCls,
    SentenceEncoderSbert,
)

BERT = "bert-base-uncased"
BERT_CLASS = BertModel
BERT_TOKENIZER_CLASS = BertTokenizerFast
D = 768
SBERT = "sentence-transformers/all-MiniLM-L6-v2"
SBERT_D = 384


class SentenceEncoderMock(SentenceEncoder):
    def forward(self, net_inputs: Dict[str, Tensor]) -> Tensor:
        """Return the feature vectors of the given inputs.

        Args:
            net_inputs (Dict[str, Tensor]): The inputs of huggingface models.

        Returns:
            Tensor: Output tensor of shape `(batch_size, embed_dim)`.
        """
        net_outputs = self.model(**net_inputs)
        return net_outputs["last_hidden_state"][:, 0]


@pytest.fixture(scope="module")
def mock_encoder():
    return SentenceEncoderMock(BERT)


@pytest.fixture(scope="module")
def avg_encoder():
    return SentenceEncoderAvg(BERT)


@pytest.fixture(scope="module")
def cls_encoder():
    return SentenceEncoderCls(BERT)


@pytest.fixture(scope="module")
def sbert_encoder():
    return SentenceEncoderSbert(SBERT)


class TestSentenceEncoder:
    def test__init__(self, mock_encoder):
        assert mock_encoder.name_or_path == BERT
        assert (
            isinstance(mock_encoder.model, BERT_CLASS)
            and mock_encoder.model.name_or_path == BERT
        )
        assert (
            isinstance(mock_encoder.tokenizer.tokenizer, BERT_TOKENIZER_CLASS)
            and mock_encoder.tokenizer.tokenizer.name_or_path == BERT
        )
        assert (
            len(list(filter(lambda p: p.requires_grad, mock_encoder.parameters()))) == 0
        )
        assert not mock_encoder.training

    def test_device(self, mock_encoder):
        assert isinstance(mock_encoder.device, torch.device)
        assert mock_encoder.device == next(mock_encoder.parameters()).device

    def test_freeze(self, mock_encoder):
        assert (
            len(list(filter(lambda p: p.requires_grad, mock_encoder.parameters()))) == 0
        )
        assert not mock_encoder.training

    @pytest.mark.parametrize("name_or_path", [BERT, SBERT])
    @pytest.mark.parametrize("representation", ["sbert", "avg", "cls", "dummy"])
    def test_build(self, name_or_path: str, representation: str):
        if representation not in {"avg", "cls", "sbert"}:
            with pytest.raises(NotImplementedError):
                SentenceEncoderMock.build(name_or_path, representation)
            return

        encoder = SentenceEncoderMock.build(name_or_path, representation)
        if representation == "sbert":
            assert isinstance(encoder, SentenceEncoderSbert)
        elif representation == "avg":
            assert isinstance(encoder, SentenceEncoderAvg)
        elif representation == "cls":
            assert isinstance(encoder, SentenceEncoderCls)

    def test_encode(self, mock_encoder):
        examples = ["I like apples.", "This is my pen.", "I like apples."]
        embeddings = mock_encoder.encode(examples)
        assert list(embeddings.shape) == [len(examples), D]
        assert not torch.allclose(embeddings[0], embeddings[1])
        torch.testing.assert_close(embeddings[0], embeddings[2])

    def test_get_embed_dim(self, mock_encoder):
        assert mock_encoder.get_embed_dim() == D


class TestSentenceEncoderAvg:
    def test_forward(self, avg_encoder: SentenceEncoderAvg):
        examples = ["I like apples.", "This is my pen.", "I like apples."]
        batch = avg_encoder.tokenizer(
            examples, return_tensors="pt", padding=True, truncation=True
        ).to(avg_encoder.device)
        embeddings = avg_encoder(batch)
        assert list(embeddings.shape) == [len(examples), D]
        assert not torch.allclose(embeddings[0], embeddings[1])
        torch.testing.assert_close(embeddings[0], embeddings[2])


class TestSentenceEncoderCls:
    def test_forward(self, cls_encoder: SentenceEncoderCls):
        examples = ["I like apples.", "This is my pen.", "I like apples."]
        batch = cls_encoder.tokenizer(
            examples, return_tensors="pt", padding=True, truncation=True
        ).to(cls_encoder.device)
        embeddings = cls_encoder(batch)
        assert list(embeddings.shape) == [len(examples), D]
        assert not torch.allclose(embeddings[0], embeddings[1])
        torch.testing.assert_close(embeddings[0], embeddings[2])


class TestSentenceEncoderSbert:
    def test_forward(self, sbert_encoder: SentenceEncoderSbert):
        examples = ["I like apples.", "This is my pen.", "I like apples."]
        batch = sbert_encoder.tokenizer(
            examples, return_tensors="pt", padding=True, truncation=True
        ).to(sbert_encoder.device)
        embeddings = sbert_encoder(batch)
        assert list(embeddings.shape) == [len(examples), SBERT_D]
        assert not torch.allclose(embeddings[0], embeddings[1])
        torch.testing.assert_close(embeddings[0], embeddings[2])

    def test_get_embed_dim(self, sbert_encoder: SentenceEncoderSbert):
        assert sbert_encoder.get_embed_dim() == SBERT_D
